
//#define DEBUG

#include <stdio.h>
#include <stdint.h>
#include <string.h>
#include <limits.h>
#include <map>
#include <list>
#include <vector>
#include <stdlib.h>
#include <assert.h>
#include <sys/types.h>
#include <regex.h>

#include <unordered_map>
#include <unordered_set>
#include <set>
#include <string>

#define CALL_FLAG 1
#define RUN_SHIFT 32
static const uint32_t CUTOFF_ADDRESS = 0xff000000U;

struct Method
{
    Method(uint32_t _start, std::string _name, uint32_t _size) :
            start(_start), name(_name), size(_size),
            call_count(0), return_count(0) {};
    virtual ~Method() {};

    uint32_t start;
    std::string name;
    uint32_t size;
    uint32_t call_count, return_count;
    std::map<const Method *, unsigned> link_count;
};


struct Tree_Node {
    Tree_Node(unsigned _address, unsigned _number) : 
            parent(0), 
            address(_address), number(_number), 
            count(1), size(4), header_flag(false) {};

    void loop_to(Tree_Node * next);
    void new_follower(Tree_Node * next);
    Tree_Node * old_follower(unsigned address);
    void check_tree(void);
    void merge(std::vector<Tree_Node *> & all_nodes,
                unsigned max_merge_size);

    std::vector<Tree_Node *> tree_children;
    Tree_Node * parent;
    std::set<Tree_Node *> loop_footer;
    unsigned address, number, count, size;
    bool header_flag;
};

void Tree_Node::loop_to(Tree_Node * next)
{
    next->loop_footer.insert(this);
    header_flag = true;
}

void Tree_Node::merge(std::vector<Tree_Node *> & all_nodes,
                unsigned max_merge_size)
{
    std::vector<Tree_Node *>::iterator tci;
    bool change = false;

    all_nodes.push_back(this);
    do {
        change = false;
        if (loop_footer.empty() && (!header_flag)
        && (tree_children.size() == 1)) {
            Tree_Node * next = tree_children[0];

            if (((address + size) == next->address)
            && (size < max_merge_size)
            && (!next->header_flag)
            && (next->loop_footer.empty())) {
                size += next->size;
                next->size = 0;
                next->tree_children.swap(tree_children);
                assert(next->tree_children.size() == 1);
                next->tree_children.pop_back();
                for (tci = tree_children.begin(); 
                            tci != tree_children.end(); tci++) {
                    assert((* tci)->parent == next);
                    (* tci)->parent = this;
                }
                //delete next;
                change = true;
            }
        } 
    } while (change);

    for (tci = tree_children.begin(); tci != tree_children.end(); tci++) {
        assert((* tci)->parent == this);
        (* tci)->merge(all_nodes, max_merge_size);
    }
}

void Tree_Node::check_tree(void)
{
    std::vector<Tree_Node *>::iterator tci;

    for (tci = tree_children.begin(); tci != tree_children.end(); tci++) {
        assert((* tci)->parent == this);
        (* tci)->check_tree();
    }
}

void Tree_Node::new_follower(Tree_Node * next)
{
    tree_children.push_back(next);
    next->parent = this;
   // printf("%x: push parent %x\n", next->address, this->address);
}

Tree_Node * Tree_Node::old_follower(unsigned address)
{
    unsigned i = tree_children.size();
    while (i != 0) {
        i--;
        Tree_Node * next = tree_children[i];
        if (next->address == address) {
            next->count ++;
            assert(next->parent == this);
            return next;
        }
    }
    return 0;
}

#if 0
Tree_Node * Tree_Node::rewind(void)
{
    Tree_Node * next = parent.back();
    parent.pop_back();

    printf("%x: pop parent %x\n", address, next->address);
    return next;
}

static void rewind(
        std::unordered_map<unsigned, Tree_Node *> & path,
        Tree_Node * current, Tree_Node * next)
{
    while(current != next) {
        printf("erasing %x\n", current->address);
        assert(path.count(current->address));
        assert(path[current->address] == current);
        path.erase(current->address);
        current = current->rewind();
        assert(current);
    }
    printf("stop erasing %x\n", current->address);
}
#endif


int main(int argc, char ** argv)
{
    if (argc != 5) {
        printf("usage: tree_get "
            "<trace_in file> <nm file> <graph out file> <spm size>\n");
        return 1;
    }
    const char * trace_in_file = argv[1];
    const char * nm_file = argv[2];
    const char * graph_out_file = argv[3];
    unsigned max_merge_size = atoi(argv[4]);

    FILE * trace_in = fopen(trace_in_file, "rb");
    if (trace_in == NULL) {
        printf("unable to open trace_in file\n");
        return 1;
    }

    FILE * nm_in = fopen(nm_file, "rt");
    if (nm_in == NULL) {
        printf("unable to open nm_in file\n");
        return 1;
    }

    FILE * graph_out = fopen(graph_out_file, "wt");
    if (graph_out == NULL) {
        printf("unable to open graph_out file\n");
        return 1;
    }


    std::unordered_map<uint32_t, Method *> methods;
    std::vector<Method *> methods_list;
    unsigned exit_at = 0;

    // scan nm file to identify functions
    char buf[BUFSIZ];
    regex_t pr;
    regmatch_t pmatch[4];

    int rc = regcomp(&pr, 
            "^([0-9a-f]{8}) ([0-9a-f]{8}|) *[tT] ([^ ]*).$", REG_EXTENDED);
    assert(rc == 0);

    while (fgets(buf, BUFSIZ - 1, nm_in)) {
        if (regexec(&pr, buf, 4, pmatch, 0) == 0) {
            assert(pmatch[1].rm_so == 0);
            assert(pmatch[2].rm_so > 0);
            assert(pmatch[3].rm_so > 0);

            buf[pmatch[1].rm_eo] = '\0';
            unsigned start = strtol(&buf[pmatch[1].rm_so], NULL, 16);

            unsigned size = 0;
            if (pmatch[2].rm_so != pmatch[2].rm_eo) {
                buf[pmatch[2].rm_eo] = '\0';
                size = strtol(&buf[pmatch[2].rm_so], NULL, 16);
            }
            buf[pmatch[3].rm_eo] = '\0';
            const char * name = &buf[pmatch[3].rm_so];

            if (start && (methods.count(start) == 0)) {
                Method * m = new Method(start, name, size);
                methods[start] = m;
                methods_list.push_back(m);

                if (strcmp(name, "exit") == 0) {
                    exit_at = start;
                }
            }
        }
    }
    regfree(&pr);

    // how many addresses in the trace?
    fseek(trace_in, 0, SEEK_END);
    long end_of_file = ftell(trace_in);
    assert(end_of_file > 0);
    assert((end_of_file & 3) == 0);
    uint64_t num_addresses = end_of_file / 4;
    uint64_t i;


    uint32_t old_address = 0;
    uint32_t address = 0;
    uint32_t call_flag = 0;
    fseek(trace_in, 0, SEEK_SET);
    fread(&address, 4, 1, trace_in);
    assert(methods.count(address));
    assert(methods[address]->name == "main");

    std::vector<std::pair<uint32_t, uint32_t> > return_stack;
    return_stack.push_back(std::pair<uint32_t, uint32_t>(0, 0));

    std::vector<Tree_Node *> all_nodes;

    Tree_Node * current;
    Tree_Node * next;
    Tree_Node * root;

    root = new Tree_Node(address, all_nodes.size());
    current = root;
    next = 0;
    all_nodes.push_back(root);

    std::vector<Tree_Node *> path;
    std::unordered_map<unsigned, unsigned> lookup;

    unsigned j;
    
    for (i = 0; i < num_addresses; i++) {
        fread(&address, 4, 1, trace_in);
        if (address >= CUTOFF_ADDRESS) continue;

        if (address == exit_at) break;

        uint32_t next_call_flag = address & CALL_FLAG;
        address &= ~3;


        if (call_flag && false) {
            // call identified
            if (!methods.count(address)) {
                fprintf(stdout, "Call to unknown function at %08x\n"
                    "Source %08x\n", address, old_address);
                exit(1);
            }

            Method * child = methods[address];
            child->call_count++;

            uint32_t parent_start = return_stack.back().second;
            if (methods.count(parent_start)) {
                Method * parent = methods[parent_start];
                parent->link_count[child]++;
            }

            return_stack.push_back(
                std::pair<uint32_t, uint32_t> (
                        old_address + 4, address));

            //printf("%x: call %s %x\n", old_address, 
            //            child->name.c_str(), address);

        } else if (return_stack.back().first == address) {
            // return identified
            uint32_t child_start = return_stack.back().second;
            Method * child = methods[child_start];

            return_stack.pop_back();
            assert(!return_stack.empty());
            assert(child);
            child->return_count++;

            // Return - requires special handling
            // We shall process the calling operation twice
            address -= 4;
            i--;
            next_call_flag = 0;
            fseek(trace_in, -4, SEEK_CUR);
            //printf("%x: return, reprocess\n", address);

            //printf("%x: return from %s\n", address, 
            //            child->name.c_str());
        }

        if (lookup.count(address)) {
            // Leaping back to an earlier point in the call path
            unsigned j = lookup[address];
            //printf("%x: lookup %u\n", address, j);
            assert(j < path.size());
            next = path[j];
            assert(next->address == address);
            assert(current != next);

            for (unsigned k = j + 1; k < path.size(); k++) {
                unsigned pba = path[k]->address;
                //printf("%x: erased %u %x\n", address, k, pba);
                assert(lookup.count(pba));
                assert(pba != address);
                assert(current != next);
                lookup.erase(pba);
            }
            path.resize(j + 1);
            assert(path.size() == (j + 1));

            //printf("%x: path size is %u\n", address, j + 1);
            //printf("%x: %x loop to %x\n", address, current->address, address);
            current->loop_to(next);
            current = next;
        } else {
            // Following an existing call path?
            j = path.size();
            //printf("%x: push %u\n", address, j);
            next = current->old_follower(address);
            if (!next) {
                // Adding to the call path
                next = new Tree_Node(address, all_nodes.size());
                current->new_follower(next);
                all_nodes.push_back(next);
            }
            assert(next->address == address);
            lookup[address] = j;
            path.push_back(next);
            current = next;
        }

        current = next;
        call_flag = next_call_flag;
        old_address = address;

    }
    printf("final address: %x\n", address);
    printf("final index: %u\n", (unsigned) ftell(trace_in));

    off_t crop_at = ftell(trace_in);
    fclose(trace_in);
    fclose(nm_in);

    if (address == exit_at) {
        truncate(trace_in_file, crop_at);
        printf("cropping\n");
    }

    // checking tree property
    unsigned footers = 0;

    assert(!all_nodes[0]->parent);
    assert(all_nodes[0]->number == 0);
    for (i = 1; i < all_nodes.size(); i++) {
        Tree_Node * n = all_nodes[i];
        assert(n->number == i);
        assert(n->parent);
        footers += n->loop_footer.size();
    }
    root->check_tree();

    // merging possible?
    all_nodes.clear();
    root->merge(all_nodes, max_merge_size);
    for (i = 0; i < all_nodes.size(); i++) {
        Tree_Node * n = all_nodes[i];
        n->number = i;
    }

#if 0
    FILE * graphviz_out = fopen("gv.gv", "wt");
    if (graphviz_out == NULL) {
        printf("unable to open graphviz_out file\n");
        return 1;
    }


    fprintf(graphviz_out, "digraph G {\n");
    fprintf(graph_out, "1024 %u W 4 20\n", (unsigned) all_nodes.size());
    for (i = 0; i < all_nodes.size(); i++) {
        Tree_Node * n = all_nodes[i];

        if (!n->size) continue;
        Tree_Node * p = n->parent;
        fprintf(graph_out, "%-8u %-8u %-8u 0\n",
            n->size, p ? p->number : 0, n->count);
        fprintf(graphviz_out, 
                "node [style=filled, color=black, fillcolor=white, "
                "label=\"%04x:%04x\"] N%u;\n",
                    n->address, n->address + n->size, n->number);
    }
    for (i = 0; i < all_nodes.size(); i++) {
        Tree_Node * n = all_nodes[i];

        if (!n->size) continue;
        Tree_Node * p = n->parent;
        if (p) {
            fprintf(graphviz_out, 
                    "N%u -> N%u [label=\"%u\"];\n",
                        p->number, n->number, n->count);
        }
    }
    fprintf(graph_out, "%u\n", footers);

    for (i = 0; i < all_nodes.size(); i++) {
        Tree_Node * n = all_nodes[i];

        if (!n->size) continue;
        std::set<Tree_Node *>::iterator lfi;
        
        for (lfi = n->loop_footer.begin();
                lfi != n->loop_footer.end(); lfi++) {

            fprintf(graph_out, "%-4u %-4u %u\n", 
                    n->number, (* lfi)->number, n->count);
            fprintf(graphviz_out, 
                    "N%u -> N%u [label=\"%u (!)\"];\n",
                        (* lfi)->number, n->number, (* lfi)->count);
        }
    }
    fprintf(graphviz_out, "}\n\n");

    fclose(graph_out);
    fclose(graphviz_out);
#endif

    for (i = 0; i < all_nodes.size(); i++) {
        Tree_Node * n = all_nodes[i];

        if (!n->size) {
        } else if (methods.count(n->address)) {
            fprintf(graph_out, "method(0x%jx, \"%s\", 0x%x, 0x%x)\n",
                (uint64_t) n->address | ((uint64_t) n->number << RUN_SHIFT), 
                methods[n->address]->name.c_str(),
                n->size,
                n->count);
        } else {
            fprintf(graph_out, "method(0x%jx, \"%x\", 0x%x, 0x%x)\n",
                (uint64_t) n->address | ((uint64_t) n->number << RUN_SHIFT), 
                n->address, 
                n->size,
                n->count);
        }
    }
    for (i = 0; i < all_nodes.size(); i++) {
        Tree_Node * n = all_nodes[i];
        Tree_Node * p = n->parent;

        if (n->size && p) {
            fprintf(graph_out, "link(0x%jx, 0x%jx, 0x%x)\n",
                (uint64_t) p->address | ((uint64_t) p->number << RUN_SHIFT), 
                (uint64_t) n->address | ((uint64_t) n->number << RUN_SHIFT), 
                n->count);
        }
    }
    for (i = 0; i < all_nodes.size(); i++) {
        Tree_Node * n = all_nodes[i];

        if (!n->size) continue;

        std::set<Tree_Node *>::iterator lfi;
        
        for (lfi = n->loop_footer.begin();
                lfi != n->loop_footer.end(); lfi++) {

            Tree_Node * p = (* lfi);
            unsigned count = p->count;

            std::vector<Tree_Node *>::iterator tci;
            for (tci = p->tree_children.begin(); 
                        tci != p->tree_children.end(); tci++) {

                assert((* tci)->count < count);
                count -= (* tci)->count;
            }

            fprintf(graph_out, "loop(0x%jx, 0x%jx, 0x%x)\n",
                (uint64_t) n->address | ((uint64_t) n->number << RUN_SHIFT), 
                (uint64_t) p->address | ((uint64_t) p->number << RUN_SHIFT), 
                count);
        }
    }
    fclose(graph_out);
}


